/*-------------------------------------------------------------------------+
| Copyright 2014 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.base.ui.annotation.labelprovider;

import org.eclipse.jface.viewers.EditingSupport;
import org.eclipse.jface.viewers.TableViewer;
import org.eclipse.swt.graphics.Color;
import org.eclipse.wb.swt.SWTResourceManager;
import org.fortiss.tooling.base.annotation.AnnotationEntry;
import org.fortiss.tooling.base.model.element.IAnnotatedSpecification;
import org.fortiss.tooling.base.model.element.IDerivedAnnotation;
import org.fortiss.tooling.base.ui.annotation.editingsupport.ComboBoxEditingSupport;
import org.fortiss.tooling.base.ui.annotation.editingsupport.IMultiValueAnnotationEditingSupport;
import org.fortiss.tooling.base.ui.annotation.view.IAnnotationViewPart;

/**
 * This class provides the text labels shown in {@link TableViewer}-based annotation views when no
 * item is selected for editing.
 * 
 * @author diewald, barner
 */
public class AnnotationLabelProvider extends LabelProviderBase {

	/** Background color for annotations that cannot be edited (and that are not derived). */
	public static final Color READ_ONLY_ANNOTATION_BACKGROUND_COLOR =
			SWTResourceManager.getColor(230, 230, 230);

	/** Background color for derived (computed) annotations. */
	private static final Color DERIVED_ANNOTATION_BACKGROUND_COLOR =
			SWTResourceManager.getColor(200, 240, 250);

	/** Type of annotation that is displayed by this {@link AnnotationLabelProvider}. */
	protected Class<? extends IAnnotatedSpecification> clazz;

	/**
	 * {@link EditingSupport} of the column to for which this {@link AnnotationLabelProvider}
	 * provides the label.
	 */
	EditingSupport editingSupport;

	/**
	 * Creates an {@link AnnotationLabelProvider} for a given annotation class.
	 * 
	 * @param clazz
	 *            Type of annotation that is displayed by this {@link AnnotationLabelProvider}.
	 * @param viewPart
	 *            view required to retrieve the selected elements
	 * @param editingSupport
	 *            {@link EditingSupport} of column for which the new {@link AnnotationLabelProvider}
	 *            should provided the label.
	 */
	public AnnotationLabelProvider(Class<? extends IAnnotatedSpecification> clazz,
			IAnnotationViewPart viewPart, EditingSupport editingSupport) {
		super(viewPart);
		this.clazz = clazz;
		this.editingSupport = editingSupport;
	}

	/** {@inheritDoc} */
	@Override
	public String getText(Object element) {
		if(element instanceof AnnotationEntry) {
			AnnotationEntry annotationEntry = (AnnotationEntry)element;

			Object value = null;
			try {
				value = annotationEntry.getSpecificationValue(clazz);
			} catch(Exception e) {
				return "";
			}

			if(editingSupport instanceof ComboBoxEditingSupport) {
				return ((ComboBoxEditingSupport)editingSupport).getLabel(element);
			} else if(editingSupport instanceof IMultiValueAnnotationEditingSupport) {
				return ((IMultiValueAnnotationEditingSupport<?>)editingSupport).getLabel(element);
			}

			return (value != null) ? value.toString() : "";
		}
		return "";
	}

	/** {@inheritDoc} */
	@Override
	public Color getBackground(Object element) {
		Color color = super.getBackground(element);
		if(element instanceof AnnotationEntry) {
			AnnotationEntry annotationEntry = (AnnotationEntry)element;
			if(annotationEntry.getModelElement().equals(viewPart.getCurrentlySelectedObject())) {
				return color;
			}

			return getAnnotatationBackgroundColor(annotationEntry, clazz, color);
		}
		return color;
	}

	/**
	 * Returns the background {@link Color} for the {@link AnnotationEntry} of the given
	 * {@code type}. The method returns dedicated colors for special cases such as read-only or
	 * "derived" annotations. Otherwise, the provided {@code defaultColor} is returned.
	 */
	public static Color getAnnotatationBackgroundColor(AnnotationEntry annotationEntry,
			Class<? extends IAnnotatedSpecification> type, Color defaultColor) {
		if(annotationEntry.canEdit(type)) {
			return defaultColor;
		}
		if(IDerivedAnnotation.class.isAssignableFrom(type) &&
				annotationEntry.getSpecificationValue(type) != null) {
			return DERIVED_ANNOTATION_BACKGROUND_COLOR;
		}
		return READ_ONLY_ANNOTATION_BACKGROUND_COLOR;
	}

	/**
	 * Returns the type (class) of the {@link IAnnotatedSpecification} that is handled by
	 * {@code this} {@link AnnotationLabelProvider}.
	 */
	public Class<? extends IAnnotatedSpecification> getAnnotationType() {
		return clazz;
	}
}
