/*-------------------------------------------------------------------------+
| Copyright 2017 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.common.ui.javafx;

import java.io.IOException;
import java.lang.reflect.InvocationTargetException;
import java.net.URL;
import java.util.ResourceBundle;

import org.apache.commons.lang3.SystemUtils;
import org.eclipse.fx.ui.workbench3.FXViewPart;
import org.eclipse.ui.part.ViewPart;

import javafx.application.Platform;
import javafx.fxml.FXMLLoader;
import javafx.fxml.Initializable;
import javafx.scene.Node;
import javafx.scene.Scene;
import javafx.scene.layout.AnchorPane;
import javafx.scene.layout.Pane;

/**
 * Base class to create eclipse (SWT-based) views using JavaFX GUI elements. This class requires
 * FXML file that defines the GUI elements of the view.
 * 
 * @author diewald
 */
public abstract class AF3FXViewPart extends FXViewPart implements Initializable {
	/** For later use, if we need to define different styles. */
	private String cssLocation;

	/** Flag that indicates if a custom style sheet has been applied. */
	private boolean isCustomStyleSheetApplied;

	/**
	 * Holds the relative location of the FXML JavaXML files. Roots are the source folders of the
	 * (plugin) project.
	 */
	private final String fXMLLocation;

	/** References the stub that defines the FXML file defining the layout and controller. */
	private Class<? extends AF3FXViewPart> viewerClass;

	/** References the root {@link Pane} to which GUI elements must be added to be displayed. */
	private Pane root;

	/**
	 * Constructor that is intended for {@link ViewPart}s which host a single view/node/control.
	 * 
	 * @param viewerClass
	 *            Class implementing the {@link AF3FXViewPart} to load resources from its containing
	 *            bundle.
	 * @param cssLocation
	 *            style sheet file path for the appearance settings relative to the source
	 *            (resource) folders. Pass {@code null} if the OS-native look shall be used.
	 * @throws Exception
	 *             if the given FXML file path is not given or if it cannot be found in any of the
	 *             source (resource) folders.
	 */
	public AF3FXViewPart(Class<? extends AF3FXViewPart> viewerClass, String cssLocation)
			throws Exception {
		// Initializing the base constructor with "null" for the FXML file indicates a
		// single-view/node/control usecase.
		this(viewerClass, null, cssLocation);
	}

	/**
	 * Constructor. Allows to pass the view's FXML definition file.
	 * 
	 * @param viewerClass
	 *            Class implementing the {@link AF3FXViewPart} to load resources from its containing
	 *            bundle.
	 * @param fXMLLocation
	 *            location of the fXML file relative to the source (resource) folders.
	 * @param cssLocation
	 *            style sheet file path for the appearance settings relative to the source
	 *            (resource) folders. Pass {@code null} if the OS-native look shall be used.
	 * @throws Exception
	 *             if the given FXML file path is not given or if it cannot be found in any of the
	 *             source (resource) folders.
	 */
	public AF3FXViewPart(Class<? extends AF3FXViewPart> viewerClass, String fXMLLocation,
			String cssLocation) throws Exception {
		Platform.setImplicitExit(false);
		this.viewerClass = viewerClass;
		this.fXMLLocation = fXMLLocation;
		this.cssLocation = cssLocation;
	}

	/** {@inheritDoc} */
	@Override
	protected Scene createFxScene() {
		FXMLLoader loader = new FXMLLoader();
		loader.setClassLoader(viewerClass.getClassLoader());
		loader.setController(this);
		if(fXMLLocation != null) {
			loader.setLocation(viewerClass.getResource(fXMLLocation));
			try {
				root = (Pane)loader.load();
			} catch(IOException e) {
				throw new RuntimeException("Cannot load the resource located at " + fXMLLocation,
						e);
			}
		} else {
			root = new AnchorPane();
			try {
				viewerClass.getMethod("initialize", URL.class, ResourceBundle.class).invoke(this,
						null, null);
			} catch(IllegalAccessException | IllegalArgumentException | InvocationTargetException |
					SecurityException e) {
				throw new RuntimeException("The initialize method of the View " +
						viewerClass.getSimpleName() + " threw an exception!", e);
			} catch(NoSuchMethodException e) {
				throw new RuntimeException("For single Controller views, an initialize method is " +
						"required that creates the control! This cannot be done in constructors " +
						"since the JavaFX toolkit is not initialized in that stage.", e);
			}
		}

		if(cssLocation != null) {
			isCustomStyleSheetApplied = applyStyleSheet(root, getClass(), cssLocation);
		}
		if(!isCustomStyleSheetApplied) {
			setNativeLook(root);
		}

		Scene scene = new Scene(root);
		return scene;
	}

	/** {@inheritDoc} */
	@Override
	protected void setFxFocus() {
		// Not needed.
	}

	/**
	 * Adds a single {@link Node} to this view. It embeds the given {@link Node} to the
	 * {@link AnchorPane} of this view and stretches it to the borders of the view.
	 * <p>
	 * NOTE: This method may only be used for single-control views, not FXML-based ones.
	 * 
	 * @param fxNode
	 *            {@link Node} to be added to the view.
	 */
	protected void setChildNodeFinal(Node fxNode) {
		if(fXMLLocation != null) {
			throw new RuntimeException("This method is intended for single-control views. It may" +
					" not be used for FXML-based views.");
		}
		if(!root.getChildren().isEmpty()) {
			throw new RuntimeException("The method setChildNodeFinal was invoked more than once." +
					" This method is intended for single-control views. If more controls" +
					"shall be added, please define an FXML layout file and call the appropriate" +
					" constructors.");
		}

		root.getChildren().add(fxNode);
		// Stretch the node to the borders of the view. In the createScene method, we a create a
		// root pane that is an AnchorPane, so we can safely use these calls here.
		AnchorPane.setTopAnchor(fxNode, 0.0);
		AnchorPane.setRightAnchor(fxNode, 0.0);
		AnchorPane.setBottomAnchor(fxNode, 0.0);
		AnchorPane.setLeftAnchor(fxNode, 0.0);
	}

	/**
	 * Sets the native look of the target platform to not break the UI experience. Most of the UI
	 * code is based on SWT which uses a platform native look. Currently, Windows 8&10, Windows 7,
	 * and MacOS X native looks are supported.
	 */
	private void setNativeLook(Pane pane) {
		if(SystemUtils.IS_OS_WINDOWS_10 || SystemUtils.IS_OS_WINDOWS_8) {
			applyStyleSheet(pane, AF3FXViewPart.class, "/jfxtras/JMetroLightTheme.css");
		} else if(SystemUtils.IS_OS_WINDOWS) {
			applyStyleSheet(pane, AF3FXViewPart.class, "/aerofx/win7.css");
		} else if(SystemUtils.IS_OS_MAC_OSX) {
			applyStyleSheet(pane, AF3FXViewPart.class, "/aquafx/aquafx.css");
		} else if(SystemUtils.IS_OS_LINUX) {
			// At the time of writing this method, no CSS is known that mimics GTK or QT. Thus, use
			// the default modena look.
		}
	}

	/**
	 * Applies the style sheet from the given location to the given {@link Pane}. The class loader
	 * of the plugin containing the style sheet must be given such that this resource can be loaded.
	 * For instance the class loader of the common.ui plugin must be given for the native looks.
	 * 
	 * @param pane
	 *            {@link Pane} to which the style sheet shall be applied.
	 * @param viewerClass
	 *            Class containing the style sheet resource.
	 * @param styleSheetLocation
	 *            relative location of the style sheet to its containing resource folder (aka source
	 *            folder).
	 * @return whether the style sheet could be applied.
	 */
	private boolean applyStyleSheet(Pane pane, Class<?> viewerClass, String styleSheetLocation) {
		if(viewerClass == null || styleSheetLocation == null) {
			return false;
		}
		URL cssURL = viewerClass.getResource(styleSheetLocation);
		if(cssURL != null) {
			pane.getStylesheets().add(cssURL.toExternalForm());
			return true;
		}
		return false;
	}
}
