/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH.
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *     Alexander Diewald (fortiss GmbH) - backport to AF3 & Integration
 *                                        improvements
 *
 *******************************************************************************/
package org.fortiss.tooling.common.ui.javafx.control.treetableview;

import javafx.beans.value.ChangeListener;
import javafx.event.Event;
import javafx.event.EventDispatchChain;
import javafx.event.EventDispatcher;
import javafx.scene.control.ContextMenu;
import javafx.scene.control.TreeCell;
import javafx.scene.control.TreeItem;
import javafx.scene.control.TreeView;
import javafx.scene.input.MouseButton;
import javafx.scene.input.MouseEvent;

/**
 * A JavaFX {@link TreeView} based on a {@link DynamicTreeContentProviderBase} and a
 * {@link DynamicTreeUIProviderBase}. This class is intended to be used as a wrapper for
 * {@link TreeView} and its rather complex expert API.
 */
public final class DynamicTreeViewer<T> extends DynamicTreeViewerBase<T> {
	/** The {@link TreeView} control to be managed. */
	private final TreeView<T> view;
	/** {@link TreeItem} constructed for the root object. */
	private final DynamicTreeItem<T> rootItem;
	/** The UI provider implementation. */
	private final DynamicTreeUIProviderBase<T> uiProvider;
	/** The selection change listener. */
	private ChangeListener<Object> selectionChangeListener = null;
	/** The double click handler. */
	private IDoubleClickHandler doubleClickHandler;

	/** Constructor. */
	public DynamicTreeViewer(TreeView<T> view, T root, boolean showRoot, int revealLevel,
			DynamicTreeContentProviderBase<T> contentProvider,
			DynamicTreeUIProviderBase<T> uiProvider) {
		super(contentProvider);
		this.uiProvider = uiProvider;
		// construct view
		this.view = view;
		rootItem = new DynamicTreeItem<T>(root, this);
		view.setRoot(rootItem);
		view.setShowRoot(showRoot);
		configureCellFactory();
		view.getSelectionModel().selectedItemProperty().addListener((obs, oldVal, newVal) -> {
			if(selectionChangeListener != null) {
				Object oVal = oldVal != null ? oldVal.getValue() : null;
				Object nVal = newVal != null ? newVal.getValue() : null;
				selectionChangeListener.changed(obs, oVal, nVal);
			}
		});
		rootItem.update();
		// expand to reveal (+1 if root node is not shown)
		expandItem(rootItem, showRoot ? revealLevel : revealLevel + 1);
	}

	/** Constructor. */
	public DynamicTreeViewer(T root, boolean showRoot, int revealLevel,
			DynamicTreeContentProviderBase<T> contentProvider,
			DynamicTreeUIProviderBase<T> uiProvider) {
		this(new TreeView<T>(), root, showRoot, revealLevel, contentProvider, uiProvider);
	}

	/** Updates the tree view. */
	public void update() {
		// wild cast works: see constructor
		DynamicTreeItem<T> rootItem = (DynamicTreeItem<T>)view.getRoot();
		rootItem.update();
	}

	/** Sets the selection of this tree view. */
	public void setSelection(Object value) {
		// wild cast works: see constructor
		DynamicTreeItem<T> rootItem = (DynamicTreeItem<T>)view.getRoot();
		DynamicTreeItem<T> item = findItem(rootItem, value);
		int row = view.getRow(item);
		view.getSelectionModel().select(row);
	}

	/** Searches the {@link DynamicTreeItem} for the given value from the root. */
	public DynamicTreeItem<T> findItem(T value) {
		return findItem(rootItem, value);
	}

	/** Searches the {@link DynamicTreeItem} for the given value. */
	public DynamicTreeItem<T> findItem(DynamicTreeItem<T> item, Object value) {
		if(item == null || value == null || item.getValue() == null) {
			return null;
		}
		if(item.getValue().equals(value)) {
			return item;
		}
		for(TreeItem<T> subItem : item.getChildren()) {
			if(subItem instanceof DynamicTreeItem) {
				DynamicTreeItem<T> i = findItem((DynamicTreeItem<T>)subItem, value);
				if(i != null) {
					return i;
				}
			}
		}
		return null;
	}

	/** Sets the selection change listener. */
	public void setSelectionListener(ChangeListener<Object> listener) {
		this.selectionChangeListener = listener;
	}

	/** Sets the double click handler. */
	public void setDoubleClickHandler(IDoubleClickHandler handler) {
		this.doubleClickHandler = handler;
	}

	/** Configures the cell factory for the viewer. */
	private void configureCellFactory() {
		view.setCellFactory(param -> {
			TreeCell<T> cell = new TreeCell<T>() {
				@Override
				public void updateItem(T item, boolean empty) {
					super.updateItem(item, empty);
					ContextMenu menu = null;
					if(!empty && item != null) {
						menu = uiProvider.createContextMenu(item);
						EventDispatcher original = this.getEventDispatcher();
						this.setEventDispatcher(new DoubleClickEventDispatcher(original));
						this.setText(uiProvider.getLabel(item));
						this.setGraphic(uiProvider.getIconNode(item));
					} else {
						this.setText(null);
						this.setGraphic(null);
					}
					this.setContextMenu(menu);
				}
			};
			return cell;
		});
	}

	/** Expands the tree to the given item. */
	public void expandItem(TreeItem<T> item) {
		while(item.getParent() != null) {
			item = item.getParent();
			item.setExpanded(true);
		}
	}

	/** Expands items up to the given level. */
	private void expandItem(TreeItem<T> parentItem, int revealLevel) {
		if(revealLevel <= 0) {
			return;
		}
		parentItem.setExpanded(true);
		revealLevel--;
		for(TreeItem<T> child : parentItem.getChildren()) {
			expandItem(child, revealLevel);
		}
	}

	/** Returns the underlying {@link TreeView}. */
	public TreeView<T> getControl() {
		return view;
	}

	/** Returns the UI provider. */
	public DynamicTreeUIProviderBase<T> getUIProvider() {
		return uiProvider;
	}

	/**
	 * Customized event dispatcher to fix double-click behavior. This prevents opening/closing of
	 * tree items with double click listeners assigned.
	 */
	private class DoubleClickEventDispatcher implements EventDispatcher {
		/** The original event dispatcher. */
		private final EventDispatcher original;

		/** Constructor. */
		public DoubleClickEventDispatcher(EventDispatcher original) {
			this.original = original;
		}

		/** {@inheritDoc} */
		@Override
		public Event dispatchEvent(Event event, EventDispatchChain tail) {
			if(event instanceof MouseEvent) {
				MouseEvent mevt = (MouseEvent)event;
				if(mevt.getButton() == MouseButton.PRIMARY && mevt.getClickCount() >= 2) {
					if(mevt.getEventType().equals(MouseEvent.MOUSE_CLICKED) &&
							doubleClickHandler != null) {
						DynamicTreeItem<T> item =
								(DynamicTreeItem<T>)view.getSelectionModel().getSelectedItem();
						if(item != null) {
							doubleClickHandler.doubleClicked(item.getValue());
						}
					}
					event.consume();
				}
			}
			return original.dispatchEvent(event, tail);
		}
	}
}
