/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.Collection;

import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.emf.common.command.CommandStackListener;
import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.data.ITopLevelElement;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.service.ICommandStackService;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IPersistencyService;

/**
 * This class implements the {@link ICommandStackService} interface. This class
 * mainly delegates its calls to the {@link IPersistencyService}.
 * 
 * @author hoelzl
 */
public class CommandStackService implements ICommandStackService, IIntrospectiveKernelService {
	/** The singleton instance. */
	private static final CommandStackService INSTANCE = new CommandStackService();

	/** Returns singleton instance of the service. */
	public static CommandStackService getInstance() {
		return INSTANCE;
	}

	/** Initializes the service. */
	public void initializeService() {
		// nothing to do
	}

	/** Starts the service. */
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Command Stack Service";
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service manages the command stack that provides the infrastructure needed to make changes to models." +
				"\nIt provides the following features:" +
				"\n - Bridge between command stack and undo/redo actions" +
				"\n - Bridge between command stack and editors" +
				"\n - Management of dirty state of models.";
		// TODO (#3336) How is nesting of commands handled by the undo/redo?
	}

	/** {@inheritDoc} */
	@Override
	public void addCommandStackListener(ITopLevelElement target, CommandStackListener listener) {
		if(target != null) {
			target.addCommandStackListener(listener);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void removeCommandStackListener(ITopLevelElement target, CommandStackListener listener) {
		if(target != null) {
			target.removeCommandStackListener(listener);
		} else {
			IPersistencyService ps = IPersistencyService.getInstance();
			for(ITopLevelElement top : ps.getTopLevelElements()) {
				top.removeCommandStackListener(listener);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public void runAsCommand(EObject target, Runnable runner) {
		IPersistencyService ps = IPersistencyService.getInstance();
		ITopLevelElement context = ps.getTopLevelElementFor(target);
		context.runAsCommand(runner);
	}

	/** {@inheritDoc} */
	@Override
	public boolean canUndo(EObject target) {
		IPersistencyService ps = IPersistencyService.getInstance();
		ITopLevelElement context = ps.getTopLevelElementFor(target);
		return context != null && context.canUndo();
	}

	/** {@inheritDoc} */
	@Override
	public boolean canRedo(EObject target) {
		IPersistencyService ps = IPersistencyService.getInstance();
		ITopLevelElement context = ps.getTopLevelElementFor(target);
		return context != null && context.canRedo();
	}

	/** {@inheritDoc} */
	@Override
	public void undo(EObject target) {
		IPersistencyService ps = IPersistencyService.getInstance();
		ps.getTopLevelElementFor(target).undo();
	}

	/** {@inheritDoc} */
	@Override
	public void redo(EObject target) {
		IPersistencyService ps = IPersistencyService.getInstance();
		ps.getTopLevelElementFor(target).redo();
	}

	/** {@inheritDoc} */
	@Override
	public boolean isDirty(EObject target) {
		IPersistencyService ps = IPersistencyService.getInstance();
		ITopLevelElement context = ps.getTopLevelElementFor(target);
		return context != null && context.isDirty();
	}

	/** {@inheritDoc} */
	@Override
	public void doSave(EObject target, IProgressMonitor monitor) {
		IPersistencyService ps = IPersistencyService.getInstance();
		ITopLevelElement context = ps.getTopLevelElementFor(target);
		if(context != null && context.isDirty()) {
			try {
				context.doSave(monitor);
			} catch(Exception e) {
				error(ToolingKernelActivator.getDefault(), "Error while saving.", e);
			}
		}
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public Collection<IIntrospectionItem> getIntrospectionItems() {
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return null;
	}
}
