/*-------------------------------------------------------------------------+
| Copyright 2011 fortiss GmbH                                              |
|                                                                          |
| Licensed under the Apache License, Version 2.0 (the "License");          |
| you may not use this file except in compliance with the License.         |
| You may obtain a copy of the License at                                  |
|                                                                          |
|    http://www.apache.org/licenses/LICENSE-2.0                            |
|                                                                          |
| Unless required by applicable law or agreed to in writing, software      |
| distributed under the License is distributed on an "AS IS" BASIS,        |
| WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. |
| See the License for the specific language governing permissions and      |
| limitations under the License.                                           |
+--------------------------------------------------------------------------*/
package org.fortiss.tooling.kernel.internal;

import static java.lang.String.format;
import static java.util.Collections.emptyList;
import static org.fortiss.tooling.kernel.utils.LoggingUtils.error;

import java.util.List;

import org.eclipse.emf.ecore.EObject;
import org.fortiss.tooling.kernel.ToolingKernelActivator;
import org.fortiss.tooling.kernel.extension.IElementCompositor;
import org.fortiss.tooling.kernel.extension.data.IElementCompositionContext;
import org.fortiss.tooling.kernel.extension.data.Prototype;
import org.fortiss.tooling.kernel.introspection.IIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectionItem;
import org.fortiss.tooling.kernel.introspection.IIntrospectiveKernelService;
import org.fortiss.tooling.kernel.introspection.items.ElementCompositorServiceIntrospectionDetailsItem;
import org.fortiss.tooling.kernel.model.ILibraryElementReference;
import org.fortiss.tooling.kernel.service.IElementCompositorService;
import org.fortiss.tooling.kernel.service.IKernelIntrospectionSystemService;
import org.fortiss.tooling.kernel.service.IPersistencyService;
import org.fortiss.tooling.kernel.service.base.EObjectAwareServiceBase;

/**
 * This class implements the {@link IElementCompositorService} interface.
 * 
 * @author hoelzl
 */
public final class ElementCompositorService
		extends EObjectAwareServiceBase<IElementCompositor<EObject>>
		implements IElementCompositorService, IIntrospectiveKernelService {
	/** The singleton instance. */
	private static final ElementCompositorService INSTANCE = new ElementCompositorService();

	/** Returns singleton instance of the service. */
	public static ElementCompositorService getInstance() {
		return INSTANCE;
	}

	/** The compositor extension point ID. */
	private static final String EXTENSION_POINT_NAME =
			"org.fortiss.tooling.kernel.modelElementCompositor";

	/** The compositor configuration element name. */
	private static final String CONFIGURATION_ELEMENT_NAME = "modelElementCompositor";

	/** The compositor class attribute name. */
	private static final String HANDLER_CLASS_ATTRIBUTE_NAME = "compositor";

	/** {@inheritDoc} */
	@Override
	public void startService() {
		IKernelIntrospectionSystemService.getInstance().registerService(this);
	}

	/** {@inheritDoc} */
	@Override
	public void registerElementCompositor(IElementCompositor<EObject> compositor,
			Class<?> modelElementClass) {
		addHandler(modelElementClass, compositor);
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionDescription() {
		return getIntrospectionLabel() +
				"\n\nThis service allows the registration of element compositors." +
				"\nElement compositors are responsible for adding a model element as the child of another model element (e.g. a sub-component " +
				"\nis added to a parent component). Most often contextual data (e.g. the mouse pointer position in a diagram) is used " +
				"\nto add specific model values during the composition procedure." +
				"\nElement composition is centralized with this service, since most often multiple UI gestures " +
				"\ncan result in the same model element composition (drag and drop vs. context menu vs. keyboard shortcut)" +
				"\n\nThe service extension point is '" + EXTENSION_POINT_NAME + "'.";
	}

	/** {@inheritDoc} */
	@Override
	public boolean canCompose(EObject container, EObject element,
			IElementCompositionContext context) {
		if(container instanceof ILibraryElementReference) {
			return false;
		}
		return findWorkingCompositor(container, element, context, false) != null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean compose(EObject container, EObject element, IElementCompositionContext context) {
		IElementCompositor<EObject> compositor =
				findWorkingCompositor(container, element, context, false);
		if(compositor == null) {
			error(ToolingKernelActivator.getDefault(),
					format("Cannot find compositor for container %s and element %s",
							container.getClass(), element.getClass()));
			return false;
		}

		// Generate IDs (for sub-model to be added)
		IPersistencyService.getInstance().getTopLevelElementFor(container).prepareIDs(element);

		// Add element to the model (might instantiate annotations for IModelElements)
		boolean rval = compositor.compose(container, element, context);

		return rval;
	}

	/** {@inheritDoc} */
	@Override
	public boolean canComposePrototype(Class<? extends EObject> containerClass,
			Prototype prototype) {
		List<IElementCompositor<EObject>> list = getRegisteredHandlers(containerClass);
		if(list == null) {
			return false;
		}
		for(IElementCompositor<EObject> compositor : list) {
			if(compositor.canComposePrototype(prototype)) {
				return true;
			}
		}
		return false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean canDecompose(EObject contained) {
		return findWorkingCompositor(contained.eContainer(), contained, null, true) != null;
	}

	/** {@inheritDoc} */
	@Override
	public boolean decompose(EObject contained) {
		return findWorkingCompositor(contained.eContainer(), contained, null, true)
				.decompose(contained);
	}

	/**
	 * Returns the first {@link IElementCompositor} which can compose the given
	 * element (or <code>null</code>).
	 */
	private IElementCompositor<EObject> findWorkingCompositor(EObject container, EObject contained,
			IElementCompositionContext context, boolean decompose) {
		if(container == null) {
			return null;
		}
		List<IElementCompositor<EObject>> list = getRegisteredHandlers(container.getClass());
		if(list == null) {
			return null;
		}
		for(IElementCompositor<EObject> compositor : list) {
			if(!decompose && compositor.canCompose(container, contained, context)) {
				return compositor;
			} else if(decompose && compositor.canDecompose(contained)) {
				return compositor;
			}
		}
		return null;
	}

	/** {@inheritDoc} */
	@Override
	protected String getExtensionPointName() {
		return EXTENSION_POINT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getConfigurationElementName() {
		return CONFIGURATION_ELEMENT_NAME;
	}

	/** {@inheritDoc} */
	@Override
	protected String getHandlerClassAttribute() {
		return HANDLER_CLASS_ATTRIBUTE_NAME;
	}

	/** {@inheritDoc} */
	@Override
	public String getIntrospectionLabel() {
		return "Element Compositor Service";
	}

	/** {@inheritDoc} */
	@Override
	public List<IIntrospectionItem> getIntrospectionItems() {
		return emptyList();
	}

	/** {@inheritDoc} */
	@Override
	public boolean showInIntrospectionNavigation() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public IIntrospectionDetailsItem getDetailsItem() {
		return new ElementCompositorServiceIntrospectionDetailsItem(handlerMap);
	}
}
