/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef;

import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;

/**
 * This class encapsulates the relative position and size changes occurring during feedback
 * operations, e.g., when the user is dragging a visual to move or resize it with a mouse gesture.
 */
public final class FeedbackChange {
	/** The empty feedback change. */
	public static final FeedbackChange EMPTY = new FeedbackChange();

	/** The relative X position change of this feedback change. */
	private final double deltaX;
	/** The relative Y position change of this feedback change. */
	private final double deltaY;
	/** The relative width change of this feedback change. */
	private final double deltaW;
	/** The relative height change of this feedback change. */
	private final double deltaH;

	/** Constructor. */
	public FeedbackChange(double dx, double dy, double dw, double dh) {
		this.deltaX = dx;
		this.deltaY = dy;
		this.deltaW = dw;
		this.deltaH = dh;
	}

	/** Constructor. */
	public FeedbackChange() {
		this(0, 0, 0, 0);
	}

	/** Returns the X position change. */
	public double getDeltaX() {
		return deltaX;
	}

	/** Returns Y position change. */
	public double getDeltaY() {
		return deltaY;
	}

	/** Returns width change. */
	public double getDeltaW() {
		return deltaW;
	}

	/** Returns the height change. */
	public double getDeltaH() {
		return deltaH;
	}

	/** Returns whether this feedback change is empty. */
	public boolean isEmpty() {
		return deltaX == 0 && deltaY == 0 && deltaW == 0 && deltaH == 0;
	}

	/** Applies this change to the given rectangle. */
	public Rectangle2D applyTo(Rectangle2D b) {
		return new Rectangle2D(b.getMinX() + deltaX, b.getMinY() + deltaY, b.getWidth() + deltaW,
				b.getHeight() + deltaH);
	}

	/** Applies the x and y change to the given point. */
	public Point2D applyToPoint(Point2D p) {
		return new Point2D(p.getX() + deltaX, p.getY() + deltaY);
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append("[").append(deltaX);
		sb.append(",").append(deltaY);
		sb.append(",").append(deltaW);
		sb.append(",").append(deltaH).append("]");
		return sb.toString();
	}

	/** Returns a {@link FeedbackChange} with a location change. */
	public static FeedbackChange locationFeedbackChange(double dx, double dy) {
		return new FeedbackChange(dx, dy, 0, 0);
	}

	/** Returns a {@link FeedbackChange} with a size change. */
	public static FeedbackChange sizeFeedbackChange(double dw, double dh) {
		return new FeedbackChange(0, 0, dw, dh);
	}
}
