/*******************************************************************************
 * Copyright (c) 2019 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.base;

import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture.MOVE;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.FeedbackChange.locationFeedbackChange;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.base.DraggingUtils.clampAndSnapMoveLocationToGridPoint;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerFeatures;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.FeedbackChange;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.change.Change;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IDragController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.controller.IKeyPressController;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IContentMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;

import javafx.geometry.Rectangle2D;
import javafx.scene.Cursor;
import javafx.scene.Node;
import javafx.scene.input.KeyCode;
import javafx.scene.input.KeyEvent;

/**
 * Controller implementation, which allows to move the underlying model element. Sub-classes must
 * implement {@link #move(FeedbackChange)} in order to change the underlying model with the
 * respective coordinate changes.
 */
public abstract class MoveControllerBase extends ControllerBase {
	/** Constructor. */
	public MoveControllerBase(IMVCBundle mvcb) {
		super(mvcb);
	}

	/** {@inheritDoc} */
	@Override
	public IDragController getDragController(EDragGesture gesture, Node node,
			DiagramCoordinate diagramLocation) {
		if(allowMove() && gesture == MOVE) {
			return moveController;
		}
		return super.getDragController(gesture, node, diagramLocation);
	}

	/** {@inheritDoc} */
	@Override
	public Cursor getCurrentCursor(Node node, DiagramCoordinate diagramLocation) {
		EDragGesture edg = getVisual().getDragGesture(node, diagramLocation);
		if(edg == EDragGesture.MOVE) {
			return allowMove() ? Cursor.MOVE : Cursor.DEFAULT;
		}
		return super.getCurrentCursor(node, diagramLocation);
	}

	/** {@inheritDoc} */
	@Override
	public IKeyPressController getKeyPressController(Node node,
			DiagramCoordinate lastMouseLocation) {
		if(allowMove() && allowKeyboardMove()) {
			return new IKeyPressController() {
				@Override
				public Change keyEvent(KeyEvent event, IMVCBundle bundle, Node source,
						DiagramCoordinate mouseLocation) {
					DiagramViewerFeatures features = getViewer().getFeatures();
					double hSpacing = features.getHorizontalSpacing();
					double vSpacing = features.getVerticalSpacing();
					FeedbackChange fbChange = null;
					KeyCode code = event.getCode();
					if(code == KeyCode.RIGHT) {
						fbChange = locationFeedbackChange(hSpacing, 0);
					} else if(code == KeyCode.LEFT) {
						fbChange = locationFeedbackChange(-hSpacing, 0);
					} else if(code == KeyCode.UP) {
						fbChange = locationFeedbackChange(0, -vSpacing);
					} else if(code == KeyCode.DOWN) {
						fbChange = locationFeedbackChange(0, vSpacing);
					} else {
						return MoveControllerBase.super.getKeyPressController(node,
								lastMouseLocation).keyEvent(event, bundle, source, mouseLocation);
					}
					if(fbChange != null && !fbChange.isEmpty()) {
						return new MoveChange(fbChange);
					}
					return null;
				}
			};
		}
		return super.getKeyPressController(node, lastMouseLocation);
	}

	/**
	 * Perform the move operation by altering the underlying model. The default does nothing.
	 * 
	 * @param deltaChange
	 *            the (possibly altered) delta the model should be moved.
	 */
	protected void move(FeedbackChange deltaChange) {
		// default does nothing
	}

	/** Returns whether the model element can be moved. */
	protected boolean allowMove() {
		return true;
	}

	/** Returns whether the model element can be moved by keyboard commands. */
	protected boolean allowKeyboardMove() {
		return true;
	}

	/**
	 * Returns whether move operations should be clamped to positive diagram area and snapped to the
	 * grid.
	 */
	protected boolean clampAndSnapMove() {
		return true;
	}

	/** Returns whether child elements should be moved also. */
	protected boolean moveChildren() {
		return true;
	}

	/** Sets the current feedback change for the given bundle and possibly its children. */
	private void setMoveFeedbackChange(FeedbackChange change, IMVCBundle bundle) {
		bundle.setFeedbackChange(change);
		if(moveChildren() && bundle instanceof IContentMVCBundle) {
			IContentMVCBundle cb = (IContentMVCBundle)bundle;
			for(IMVCBundle ch : cb.getChildren()) {
				setMoveFeedbackChange(change, ch);
			}
		}
	}

	/**
	 * Alters the move feedback to apply some constraints to it (e.g. no negative result location).
	 */
	protected FeedbackChange alterMoveFeedback(FeedbackChange change) {
		if(!clampAndSnapMove()) {
			return change;
		}
		Rectangle2D modelBounds = getVisual().getModelBounds();
		return clampAndSnapMoveLocationToGridPoint(modelBounds.getMinX(), modelBounds.getMinY(),
				change.getDeltaX(), change.getDeltaY(), getViewer().getFeatures());
	}

	/** The drag controller managing the move operation. */
	private final IDragController moveController = new DragControllerBase() {
		@Override
		protected final void safeDragInProgress(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			FeedbackChange altered = alterMoveFeedback(getDragDelta(diagramLocation));
			setMoveFeedbackChange(altered, startBundle);
		}

		@Override
		protected final Change safeDragCompleted(IMVCBundle bundle, Node node,
				DiagramCoordinate diagramLocation) {
			setMoveFeedbackChange(FeedbackChange.EMPTY, startBundle);
			final FeedbackChange change = alterMoveFeedback(getDragDelta(diagramLocation));
			if(change.isEmpty()) {
				return null;
			}
			return new MoveChange(change);
		}
	};

	/** A {@link Change} that calls {@link #move(FeedbackChange)}. */
	private final class MoveChange implements Change {
		/** The feedback change to be applied. */
		private final FeedbackChange change;

		/** Constructor. */
		public MoveChange(FeedbackChange change) {
			this.change = change;
		}

		/** {@inheritDoc} */
		@Override
		public void applyChange() {
			move(change);
		}
	}
}
