/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.base;

import static java.lang.Math.abs;
import static java.lang.Math.atan2;
import static java.lang.Math.cos;
import static java.lang.Math.sin;
import static javafx.scene.paint.Color.RED;
import static javafx.scene.paint.Color.TRANSPARENT;
import static javafx.scene.paint.Color.color;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.FeedbackChange;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.ILinkVisual;

import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;
import javafx.scene.shape.Rectangle;

/** Base implementation for {@link ILinkVisual}. */
public abstract class LinkVisualBase extends MVCBundlePartWithEffectsBase implements ILinkVisual {
	/** Stores the feedback change for each bend point. */
	protected final Map<Integer, FeedbackChange> feedbackChanges = new HashMap<>();
	/** The selection feedback handle rectangle for the start of the line link. */
	protected final Rectangle startFeedbackHandle;
	/** The selection feedback handle rectangle for the end of the line link. */
	protected final Rectangle endFeedbackHandle;

	/** Constructor. */
	public LinkVisualBase(ILinkMVCBundle mvcb) {
		super(mvcb);
		double feedbackSize = getFeedbackMarkerSize();
		double fs2 = feedbackSize / 2;
		Point2D startAnchorPoint = getStartAnchorPoint();
		startFeedbackHandle = new Rectangle(startAnchorPoint.getX() - fs2,
				startAnchorPoint.getY() - fs2, feedbackSize, feedbackSize);
		startFeedbackHandle.setFill(TRANSPARENT);
		startFeedbackHandle.setStroke(RED);
		Point2D endAnchorPoint = getEndAnchorPoint();
		endFeedbackHandle = new Rectangle(endAnchorPoint.getX() - fs2, endAnchorPoint.getY() - fs2,
				feedbackSize, feedbackSize);
		endFeedbackHandle.setFill(TRANSPARENT);
		endFeedbackHandle.setStroke(RED);
	}

	/** {@inheritDoc} */
	@Override
	public final Rectangle2D getModelBounds() {
		// links do not need model bounds
		return null;
	}

	/** {@inheritDoc} */
	@Override
	public void setFeedbackChangeForBendPoint(int index, FeedbackChange change) {
		if(change == null) {
			feedbackChanges.remove(index);
		} else {
			feedbackChanges.put(index, change);
		}
		getViewer().updateVisual(getMVCBundle());
	}

	/**
	 * Returns the current {@link FeedbackChange} for the bend-point at the given index. If there is
	 * no entry at the given index then {@link FeedbackChange#EMPTY} is returned.
	 */
	protected FeedbackChange getFeedbackChangeForBendPoint(int index) {
		FeedbackChange result = feedbackChanges.get(index);
		if(result == null) {
			return FeedbackChange.EMPTY;
		}
		return result;
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getLinkAnchorage(DiagramCoordinate indication) {
		return indication;
	}

	/** Returns the color to be used when the connection is highlighted as incoming link. */
	protected Paint getHighlightIncomingLinkColor() {
		return Color.GREENYELLOW;
	}

	/** Returns the color to be used when the connection is highlighted as outgoing link. */
	protected Paint getHighlightOutgoingLinkColor() {
		return Color.TOMATO;
	}

	/**
	 * Returns the point on a circle centered in the current bounds intersecting with a line between
	 * the center point and the given point.
	 */
	protected DiagramCoordinate getCircleLocation(Rectangle2D bounds, Point2D target) {
		double sx = bounds.getMinX() + bounds.getWidth() / 2;
		double sy = bounds.getMinY() + bounds.getHeight() / 2;

		double a = atan2(target.getY() - sy, target.getX() - sx);
		double nx = sx + cos(a) * (bounds.getWidth() / 2 - 1);
		double ny = sy + sin(a) * (bounds.getHeight() / 2 - 1);
		return new DiagramCoordinate(nx, ny);
	}

	/** Returns the middle point of the rectangle border closest to the target point. */
	protected final DiagramCoordinate getRectangleLocation(Rectangle2D ib, Point2D target) {
		double cx = ib.getMinX() + ib.getWidth() / 2;
		double cy = ib.getMinY() + ib.getHeight() / 2;
		double dx = cx - target.getX();
		double dy = cy - target.getY();
		if(abs(dx) > abs(dy)) {
			if(cx < target.getX()) {
				return new DiagramCoordinate(ib.getMaxX(), cy);
			}
			return new DiagramCoordinate(ib.getMinX(), cy);
		}
		if(cy < target.getY()) {
			return new DiagramCoordinate(cx, ib.getMaxY());
		}
		return new DiagramCoordinate(cx, ib.getMinY());
	}

	/** The interaction shading color. */
	private static final Color INTERACTION_SHADING = color(0, 0, 0, .1);

	/** Returns the color for the interaction area. */
	protected final Paint getInteractionColor() {
		return getViewer().getFeatures().isInteractionAreaShadingEnabled() ? INTERACTION_SHADING
				: Color.TRANSPARENT;
	}

	/** Returns the size of the feedback markers. */
	protected double getFeedbackMarkerSize() {
		return getMVCBundle().getViewer().getFeatures().getHorizontalSpacing() / 2;
	}

	/** {@inheritDoc} */
	@Override
	public int getBendPointIndex(Node node) {
		if(node == startFeedbackHandle) {
			return START_OF_LINK_BEND_POINT_INDEX;
		}
		if(node == endFeedbackHandle) {
			return END_OF_LINK_BEND_POINT_INDEX;
		}
		return NO_BEND_POINT_INDEX;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isBendPointHandle(Node node) {
		return node == startFeedbackHandle || node == endFeedbackHandle;
	}

	/** {@inheritDoc} */
	@Override
	public void disableFeedback() {
		startFeedbackHandle.setMouseTransparent(true);
		endFeedbackHandle.setMouseTransparent(true);
	}

	/** {@inheritDoc} */
	@Override
	public void enableFeedback() {
		startFeedbackHandle.setMouseTransparent(false);
		endFeedbackHandle.setMouseTransparent(false);
	}
}
