/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH.
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.base;

import static javafx.scene.paint.Color.CORNFLOWERBLUE;
import static javafx.scene.paint.Color.DEEPSKYBLUE;
import static javafx.scene.paint.Color.ORANGE;
import static javafx.scene.paint.Color.RED;
import static javafx.scene.paint.Color.TRANSPARENT;
import static javafx.scene.paint.Color.color;
import static javafx.scene.shape.StrokeType.INSIDE;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture.MOVE;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture.NEW_LINK;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture.NONE;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture.RESIZE_H;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture.RESIZE_V;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture.RESIZE_VH;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramLayers;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramLayers.ILayer;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.IVisual;

import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.geometry.VPos;
import javafx.scene.Node;
import javafx.scene.effect.DropShadow;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;
import javafx.scene.shape.Rectangle;
import javafx.scene.shape.Shape;
import javafx.scene.shape.StrokeType;
import javafx.scene.text.Text;
import javafx.scene.text.TextAlignment;

/**
 * Base class implementing the {@link IVisual} interface.
 * <P>
 * This implementation provides a simple {@link DropShadow} hover effect and a hover label.
 */
public abstract class VisualBase extends MVCBundlePartWithEffectsBase implements IVisual {
	/** The visible shape of this visual. */
	private final Shape visualShape;
	/** The invisible hit area of this visual. */
	private final Shape hitAreaShape;
	/** The selection rectangle. */
	protected final Rectangle selectionRectangle = new Rectangle();
	/** The hover text label. */
	private final Text hoverText = new Text("");

	/** Constructor. */
	public VisualBase(IMVCBundle mvcb, Shape visual, Shape hit) {
		super(mvcb);
		this.visualShape = visual;
		this.hitAreaShape = hit;
		hitAreaShape.setFill(TRANSPARENT);
		hitAreaShape.setStroke(TRANSPARENT);
		selectionRectangle.setFill(TRANSPARENT);
		selectionRectangle.setStrokeWidth(2);
		hoverText.setMouseTransparent(true);
	}

	/** Returns the hit area shape. */
	protected final Shape getHitAreaShape() {
		return hitAreaShape;
	}

	/** Returns visual shape. */
	protected final Shape getVisualShape() {
		return visualShape;
	}

	/** {@inheritDoc} */
	@Override
	protected void createHoverEffect(DiagramLayers layers) {
		Color shadow = getHoverShadowColor();
		if(shadow != null && visualShape != null) {
			visualShape.setEffect(new DropShadow(10, 3, 3, shadow));
		}
		// handle hover text
		String text = getHoverText();
		if(text == null) {
			return;
		}
		hoverText.setText(text);
		Point2D loc = getHoverTextLocation();
		Rectangle2D bounds = getCurrentBounds();
		double tx = bounds.getMinX() + loc.getX();
		hoverText.setX(tx);
		double ty = bounds.getMinY() + loc.getY();
		hoverText.setY(ty);
		hoverText.setTextAlignment(TextAlignment.LEFT);
		hoverText.setTextOrigin(VPos.TOP);
		hoverText.setMouseTransparent(true);
		layers.getVisualFeedbackLayer().add(hoverText, getMVCBundle());
	}

	/** {@inheritDoc} */
	@Override
	protected void removeHoverEffect(DiagramLayers layers) {
		if(visualShape != null) {
			visualShape.setEffect(null);
		}
		layers.getVisualFeedbackLayer().remove(hoverText);
	}

	/** {@inheritDoc} */
	@Override
	public void removeAllVisuals(DiagramLayers layers) {
		super.removeAllVisuals(layers);

		getLayerForHitShape(layers).remove(hitAreaShape);
		getLayerForVisualShape(layers).remove(visualShape);
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);

		IMVCBundle bundle = getMVCBundle();
		if(visualShape != null) {
			if(enableVisual()) {
				if(visualShape.getParent() == null) {
					getLayerForVisualShape(layers).add(visualShape, bundle);
				}
			} else {
				getLayerForVisualShape(layers).remove(visualShape);
			}
		}
		if(hitAreaShape != null) {
			if(enableInteraction()) {
				if(hitAreaShape.getParent() == null) {
					getLayerForHitShape(layers).add(hitAreaShape, bundle);
				}
			} else {
				getLayerForHitShape(layers).remove(hitAreaShape);
			}
		}
	}

	/**
	 * Returns the layer to be used for the {@link #getVisualShape() visual shape}. The default is
	 * {@link DiagramLayers#getContentLayer()}
	 */
	protected ILayer getLayerForVisualShape(DiagramLayers layers) {
		return layers.getContentLayer();
	}

	/**
	 * Returns the layer to be used for the {@link #getHitAreaShape() hit area shape}. The default
	 * is {@link DiagramLayers#getContentInteractionLayer()}
	 */
	protected ILayer getLayerForHitShape(DiagramLayers layers) {
		return layers.getContentInteractionLayer();
	}

	/** {@inheritDoc} */
	@Override
	public boolean enableVisual() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean enableInteraction() {
		return true;
	}

	/** Updates and adds the selection feedback rectangle. */
	protected void updateSelectionFeedbackRectangle(Paint color, DiagramLayers layers) {
		Rectangle2D bounds = getCurrentBounds();
		if(bounds == null) {
			return;
		}
		selectionRectangle.setX(bounds.getMinX());
		selectionRectangle.setY(bounds.getMinY());
		selectionRectangle.setWidth(bounds.getWidth());
		selectionRectangle.setHeight(bounds.getHeight());
		selectionRectangle.setStroke(color);
		layers.getVisualFeedbackLayer().add(selectionRectangle, getMVCBundle());
	}

	/** {@inheritDoc} */
	@Override
	protected void applySelectedNotFocusedEffect(DiagramLayers layers) {
		updateSelectionFeedbackRectangle(getSelectedNotFocusedColor(), layers);
	}

	/** {@inheritDoc} */
	@Override
	protected void applySecondarySelectedEffect(DiagramLayers layers) {
		updateSelectionFeedbackRectangle(getSecondarySelectedColor(), layers);
	}

	/** {@inheritDoc} */
	@Override
	protected void applySelectedFocusedEffect(DiagramLayers layers) {
		updateSelectionFeedbackRectangle(getSelectedFocusedColor(), layers);
	}

	/** {@inheritDoc} */
	@Override
	protected void applyNotSelectedEffect(DiagramLayers layers) {
		super.applyNotSelectedEffect(layers);
		layers.getVisualFeedbackLayer().remove(selectionRectangle);
	}

	/** {@inheritDoc} */
	@Override
	public EDragGesture getDragGesture(Node node, DiagramCoordinate diagramLocation) {
		boolean notSelected = !getViewer().isPrimarySelected(getMVCBundle());
		EDragGesture gesture = dragGestureHitTest(node, diagramLocation);
		if(gesture == NEW_LINK && requireSelectionForNewLinkGesture() && notSelected) {
			// override when not selected
			return NONE;
		}
		if(gesture == MOVE && requireSelectionForMoveGesture() && notSelected) {
			// override when not selected
			return NONE;
		}
		if(gesture == RESIZE_H || gesture == RESIZE_V || gesture == RESIZE_VH) {
			if(requireSelectionForResizeGesture() && notSelected) {
				// override when not selected
				return NONE;
			}
		}
		return gesture;
	}

	/**
	 * Performs a hit test on the given node returning the expected gesture. Note that the result
	 * may be overridden if the node is not selected and
	 * {@link #requireSelectionForNewLinkGesture()} or {@link #requireSelectionForMoveGesture()}
	 * return {@code true}, respectively. The default returns {@link EDragGesture#NONE}.
	 * 
	 * @param node
	 *            the node currently under the mouse pointer
	 * @param diagramLocation
	 *            the location in diagram coordinates
	 * @return the drag gesture
	 */
	protected EDragGesture dragGestureHitTest(Node node, DiagramCoordinate diagramLocation) {
		return NONE;
	}

	/**
	 * Returns whether the visual must be selected when creating links is allowed. The default is
	 * {@code false}.
	 */
	protected boolean requireSelectionForNewLinkGesture() {
		return false;
	}

	/**
	 * Returns whether the visual must be selected when moving is allowed. The default is
	 * {@code true}.
	 */
	protected boolean requireSelectionForMoveGesture() {
		return true;
	}

	/**
	 * Returns whether the visual must be selected when resizing is allowed. The default is
	 * {@code true}.
	 */
	protected boolean requireSelectionForResizeGesture() {
		return true;
	}

	/** Returns the opacity of the content visual. */
	protected double getOpacity() {
		return 0.8;
	}

	/** Returns the color used to paint the border of the visual. */
	protected Paint getBorderColor() {
		return CORNFLOWERBLUE;
	}

	/** Returns the width of the border. */
	protected double getBorderWidth() {
		return 2;
	}

	/** Returns the type of the border. */
	protected StrokeType getBorderType() {
		return INSIDE;
	}

	/** Returns the color for the interior of the visual. */
	protected Paint getFillColor() {
		return DEEPSKYBLUE;
	}

	/** Returns the color of the selection rectangle if it also has focus. */
	protected Color getSelectedFocusedColor() {
		return RED;
	}

	/** Returns the color of the selection rectangle if it part of a secondary selection. */
	protected Color getSecondarySelectedColor() {
		return ORANGE;
	}

	/** Returns the color of the selection rectangle if it does not have focus. */
	protected Color getSelectedNotFocusedColor() {
		return RED;
	}

	/** Returns the size of the hit area as an outset of the visible rectangle. */
	protected double getHitAreaOutset() {
		return 10;
	}

	/** The interaction shading color. */
	protected static final Color INTERACTION_SHADING = color(0, 0, 0, .1);

	/** Returns the color for the interaction area. */
	protected Color getInteractionColor() {
		return getViewer().getFeatures().isInteractionAreaShadingEnabled() ? INTERACTION_SHADING
				: Color.TRANSPARENT;
	}

	/** Returns the color of the hover drop shadow. */
	protected Color getHoverShadowColor() {
		return Color.DARKGRAY.darker();
	}

	/** Returns the hover text. */
	protected String getHoverText() {
		return null;
	}

	/** Returns the hover text location . */
	protected Point2D getHoverTextLocation() {
		return new Point2D(0, getCurrentBounds().getHeight());
	}

	/** Returns the center point of the current bounds. */
	protected final DiagramCoordinate getCurrentBoundsCenter() {
		Rectangle2D b = getVisual().getCurrentBounds();
		double x = b.getMinX() + b.getWidth() / 2;
		double y = b.getMinY() + b.getHeight() / 2;
		return new DiagramCoordinate(x, y);
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getLinkAnchorage(DiagramCoordinate indication) {
		return getCurrentBoundsCenter();
	}

	/** {@inheritDoc} */
	@Override
	public void requestFocus() {
		getVisualShape().requestFocus();
	}
}
