/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.elliptic;

import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.*;

import java.util.LinkedList;
import java.util.List;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramLayers;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramLayers.ILayer;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.FeedbackChange;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IAnchorageMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.impl.LinkMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.base.LinkVisualBase;

import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;

/** Base class for {@link LinkVisualBase link visuals} with Bezier curve lines. */
public abstract class CurveLinkVisualBase extends LinkVisualBase {
	/** Stores the visual segments. */
	private final List<CurveSegment> segments = new LinkedList<>();
	/** Flag storing if the lines have been added. */
	private boolean curvesAddedToSceneGraph = false;

	/** Constructor. */
	public CurveLinkVisualBase(ILinkMVCBundle mvcb) {
		super(mvcb);
	}

	/** Adds the curve segments to the scene graph. */
	private void addCurveSegments(DiagramLayers layers) {
		LinkMVCBundle linkBundle = getLinkBundle();
		double markerSize = getFeedbackMarkerSize();
		double ms2 = markerSize / 2;
		// start point
		Point2D sp = getFeedbackChangeForBendPoint(START_OF_LINK_BEND_POINT_INDEX)
				.applyToPoint(getStartAnchorPoint());
		Point2D cp1 = getBendPointLocation(0);
		Point2D cp2 = getBendPointLocation(1);
		startFeedbackHandle.setX(sp.getX() - ms2);
		startFeedbackHandle.setY(sp.getY() - ms2);
		startFeedbackHandle.setWidth(markerSize);
		startFeedbackHandle.setHeight(markerSize);
		// handle bend points
		int pts = getNumberOfBendPoints();
		for(int i = 2; i < pts; i += 3) {
			Point2D ep = getBendPointLocation(i);
			makeCurve(sp, cp1, cp2, ep, getInvisibleSelectionLineWidth(), markerSize, markerSize,
					false, true, 0);
			sp = ep;
			cp1 = getBendPointLocation(i + 1);
			cp2 = getBendPointLocation(i + 2);
			// bpModel = getBendPointModel();
		}
		// end point
		Point2D ep = getFeedbackChangeForBendPoint(END_OF_LINK_BEND_POINT_INDEX)
				.applyToPoint(getEndAnchorPoint());
		// Last bend point is not selectable => set its feedback rectangle size to 0
		makeCurve(sp, cp1, cp2, ep, getInvisibleSelectionLineWidth(), 0, markerSize,
				showArrowOnLastSegment(), useLineArrow(), getArrowLength());
		endFeedbackHandle.setX(ep.getX() - ms2);
		endFeedbackHandle.setY(ep.getY() - ms2);
		endFeedbackHandle.setWidth(markerSize);
		endFeedbackHandle.setHeight(markerSize);

		for(CurveSegment cs : segments) {
			cs.addLinkNodes(layers, linkBundle);
		}
		curvesAddedToSceneGraph = true;
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);
		// if no longer enabled => remove
		if(!enableVisual()) {
			if(curvesAddedToSceneGraph) {
				removeCurveSegments(layers);
			}
			return;
		}
		// check for model updates (e.g new bend-point)
		int pts = getNumberOfBendPoints();
		int expectedSegments = 1 + (pts - 2) / 3;
		if(!curvesAddedToSceneGraph) {
			addCurveSegments(layers);
		}
		if(segments.size() != expectedSegments) {
			removeCurveSegments(layers);
			addCurveSegments(layers);
			if(getViewer().getSelection().getPrimarySelection() == getLinkBundle()) {
				if(getViewer().hasFocus()) {
					applySelectedFocusedEffect(layers);
				} else {
					applySelectedNotFocusedEffect(layers);
				}
			}
			return;
		}

		int segIndex = 0;
		double markerSize = getFeedbackMarkerSize();
		double ms2 = markerSize / 2;
		// start point
		Point2D sp = getFeedbackChangeForBendPoint(START_OF_LINK_BEND_POINT_INDEX)
				.applyToPoint(getStartAnchorPoint());
		startFeedbackHandle.setX(sp.getX() - ms2);
		startFeedbackHandle.setY(sp.getY() - ms2);
		startFeedbackHandle.setWidth(markerSize);
		startFeedbackHandle.setHeight(markerSize);
		Point2D cp1 = getFeedbackChangeForBendPoint(0).applyToPoint(getBendPointLocation(0));
		Point2D cp2 = getFeedbackChangeForBendPoint(1).applyToPoint(getBendPointLocation(1));
		// handle bend points
		for(int i = 2; i < pts; i += 3) {
			Point2D ep = getFeedbackChangeForBendPoint(i).applyToPoint(getBendPointLocation(i));
			segments.get(segIndex).update(sp, cp1, cp2, ep, markerSize, markerSize, 0);
			sp = ep;
			cp1 = getFeedbackChangeForBendPoint(i + 1).applyToPoint(getBendPointLocation(i + 1));
			cp2 = getFeedbackChangeForBendPoint(i + 2).applyToPoint(getBendPointLocation(i + 2));
			segIndex++;
		}
		// end point
		Point2D ep = getFeedbackChangeForBendPoint(END_OF_LINK_BEND_POINT_INDEX)
				.applyToPoint(getEndAnchorPoint());
		segments.get(segIndex).update(sp, cp1, cp2, ep, 0, markerSize, getArrowLength());
		endFeedbackHandle.setX(ep.getX() - ms2);
		endFeedbackHandle.setY(ep.getY() - ms2);
		endFeedbackHandle.setWidth(markerSize);
		endFeedbackHandle.setHeight(markerSize);

		// coloring
		Paint color = getLineColor();
		IMVCBundle mvcBundle = getMVCBundle();
		if(mvcBundle.hasTag(HIGHLIGHT_INCOMING_LINK_TAG)) {
			color = getHighlightIncomingLinkColor();
		} else if(mvcBundle.hasTag(HIGHLIGHT_OUTGOING_LINK_TAG)) {
			color = getHighlightOutgoingLinkColor();
		} else if(mvcBundle.hasTag(PRIMARY_SELECTION_TAG)) {
			if(mvcBundle.hasTag(FOCUS_TAG)) {
				color = getSelectionFocusedColor();
			} else {
				color = getSelectionNotFocusedColor();
			}
		}
		for(CurveSegment seg : segments) {
			seg.setStrokeColor(color);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void removeAllVisuals(DiagramLayers layers) {
		super.removeAllVisuals(layers);
		removeCurveSegments(layers);
	}

	/** Removes the curve segments from the scene graph. */
	private void removeCurveSegments(DiagramLayers layers) {
		for(CurveSegment cs : segments) {
			cs.removeLinkNodes(layers);
			cs.removeFeedbackNodes(layers);
		}
		ILayer l = layers.getLinkInteractionLayer();
		l.remove(startFeedbackHandle);
		l.remove(endFeedbackHandle);
		segments.clear();
		curvesAddedToSceneGraph = false;
	}

	/** {@inheritDoc} */
	@Override
	public boolean enableVisual() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public boolean enableInteraction() {
		return true;
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getStartAnchorPoint() {
		IAnchorageMVCBundle start = getLinkBundle().getStartAnchorage();
		Rectangle2D sb = start.getVisual().getCurrentBounds();

		Point2D target;
		int pts = getNumberOfBendPoints();
		if(pts == 0) {
			IAnchorageMVCBundle end = getLinkBundle().getEndAnchorage();
			Rectangle2D eb = end.getVisual().getCurrentBounds();
			target = new Point2D(eb.getMinX() + eb.getWidth() / 2,
					eb.getMinY() + eb.getHeight() / 2);
		} else {
			target = getCurrentBendPointLocation(0);
		}
		return getStartAnchorLocation(sb, target);
	}

	/** {@inheritDoc} */
	@Override
	public DiagramCoordinate getEndAnchorPoint() {
		IAnchorageMVCBundle end = getLinkBundle().getEndAnchorage();
		Rectangle2D eb = end.getVisual().getCurrentBounds();

		Point2D target;
		int pts = getNumberOfBendPoints();
		if(pts == 0) {
			IAnchorageMVCBundle start = getLinkBundle().getStartAnchorage();
			Rectangle2D sb = start.getVisual().getCurrentBounds();
			target = new Point2D(sb.getMinX() + sb.getWidth() / 2,
					sb.getMinY() + sb.getHeight() / 2);
		} else {
			target = getCurrentBendPointLocation(pts - 1);
		}
		return getEndAnchorLocation(eb, target);
	}

	/** {@inheritDoc} */
	@Override
	public void applySelectedFocusedEffect(DiagramLayers layers) {
		ILinkMVCBundle bundle = getLinkBundle();
		Paint selectionFocusedColor = getSelectionFocusedColor();
		Paint interactionColor = getInteractionColor();
		for(CurveSegment seg : segments) {
			seg.setStrokeColor(selectionFocusedColor);
			seg.setClickableLinesColor(interactionColor);
			seg.addFeedbackNodes(layers, bundle);
		}
		ILayer linkInteractionLayer = layers.getLinkInteractionLayer();
		linkInteractionLayer.add(startFeedbackHandle, bundle);
		linkInteractionLayer.add(endFeedbackHandle, bundle);
	}

	/** {@inheritDoc} */
	@Override
	public void applySelectedNotFocusedEffect(DiagramLayers layers) {
		removeFeedback(layers, getSelectionNotFocusedColor());
	}

	/** {@inheritDoc} */
	@Override
	protected void applyNotSelectedEffect(DiagramLayers layers) {
		removeFeedback(layers, getInteractionColor());
	}

	/** Removes the feedback handles and sets the link to the given color. */
	private void removeFeedback(DiagramLayers layers, Paint lineColor) {
		Paint selectionNotFocusedColor = lineColor;
		Paint interactionColor = getInteractionColor();
		for(CurveSegment seg : segments) {
			seg.setStrokeColor(selectionNotFocusedColor);
			seg.setClickableLinesColor(interactionColor);
			seg.removeFeedbackNodes(layers);
		}
		ILayer linkInteractionLayer = layers.getLinkInteractionLayer();
		linkInteractionLayer.remove(startFeedbackHandle);
		linkInteractionLayer.remove(endFeedbackHandle);
	}

	/**
	 * Returns the anchor location w.r.t. the interface node bonds and the given target location
	 * (e.g. the first bend-point or the end interface node center point).
	 */
	protected abstract DiagramCoordinate getStartAnchorLocation(Rectangle2D ifcBounds,
			Point2D target);

	/**
	 * Returns the anchor location w.r.t. the interface node bonds and the given target location
	 * (e.g. the last bend-point or the start interface node center point).
	 */
	protected abstract DiagramCoordinate getEndAnchorLocation(Rectangle2D ifcBounds,
			Point2D target);

	/** Returns the color to be used when the connection is selected. */
	protected Paint getSelectionFocusedColor() {
		return Color.RED;
	}

	/** Returns the color to be used when the connection is selected, but not focused. */
	protected Paint getSelectionNotFocusedColor() {
		return Color.ORANGE;
	}

	/** Returns the color to be used when the connection is not selected. */
	protected Paint getLineColor() {
		return Color.BLACK;
	}

	/**
	 * Returns the point of the bend-point at the given index altered by its current feedback
	 * change.
	 */
	private Point2D getCurrentBendPointLocation(int bendPointIndex) {
		if(bendPointIndex >= 0 && bendPointIndex < getNumberOfBendPoints()) {
			Point2D bendPoint = getBendPointLocation(bendPointIndex);
			FeedbackChange chg = getFeedbackChangeForBendPoint(bendPointIndex);
			double x = bendPoint.getX() + chg.getDeltaX();
			double y = bendPoint.getY() + chg.getDeltaY();
			return new Point2D(x, y);
		}
		throw new IndexOutOfBoundsException("No bend-point exists at index " + bendPointIndex);
	}

	/**
	 * Returns the model of the bend-point at the given index. Each bend-point must be identified by
	 * a unique object, which must not change as long as the bend-point exists.
	 */
	protected abstract Object getBendPointModel(int i);

	/** Returns the number of bend-points. */
	protected abstract int getNumberOfBendPoints();

	/** Returns the width of the invisible selection line. */
	protected abstract double getInvisibleSelectionLineWidth();

	/** Creates the curve segment. */
	private void makeCurve(Point2D sp, Point2D cp1, Point2D cp2, Point2D ep, double selWidth,
			double bpMarkerSize, double cpMarkerSize, boolean showArrow, boolean useLineArrow,
			double arrowLength) {
		CurveSegment segVis = new CurveSegment(sp.getX(), sp.getY(), cp1.getX(), cp1.getY(),
				cp2.getX(), cp2.getY(), ep.getX(), ep.getY(), selWidth, bpMarkerSize, cpMarkerSize,
				showArrow, useLineArrow, arrowLength);
		segments.add(segVis);
	}

	/** Returns the {@link LinkMVCBundle}. */
	private LinkMVCBundle getLinkBundle() {
		return (LinkMVCBundle)getMVCBundle();
	}

	/** {@inheritDoc} */
	@Override
	public Rectangle2D getCurrentBounds() {
		// connections do not need bounds
		return Rectangle2D.EMPTY;
	}

	/** {@inheritDoc} */
	@Override
	public int getBendPointIndex(Node node) {
		int superResult = super.getBendPointIndex(node);
		if(superResult != NO_BEND_POINT_INDEX) {
			return superResult;
		}
		for(int i = 0; i < segments.size(); i++) {
			CurveSegment cs = segments.get(i);
			if(node == cs.getFirstControlPointHandle()) {
				return i * 3;
			} else if(node == cs.getSecondControlPointHandle()) {
				return 1 + i * 3;
			} else if(node == cs.getBendPointHandle()) {
				return 2 + i * 3;
			} else if(node == cs.getVisibleCurve() || node == cs.getClickableCurve()) {
				return 2 + i * 3;
			}
		}
		return NO_BEND_POINT_INDEX;
	}

	/** {@inheritDoc} */
	@Override
	public boolean isBendPointHandle(Node node) {
		for(CurveSegment cs : segments) {
			if(node == cs.getBendPointHandle() || node == cs.getFirstControlPointHandle() ||
					node == cs.getSecondControlPointHandle()) {
				return true;
			}
		}
		return super.isBendPointHandle(node);
	}

	/** {@inheritDoc} */
	@Override
	public void requestFocus() {
		if(!segments.isEmpty()) {
			// segment list was tested for emptiness => get(0) works
			segments.get(0).getVisibleCurve().requestFocus();
		}
	}

	/** Returns whether the last line segment should show an arrow head marker. */
	protected abstract boolean showArrowOnLastSegment();

	/** Returns the type of the arrow to be used (line vs. filled). The default is filled. */
	protected boolean useLineArrow() {
		return false;
	}

	/** Returns the length of the arrow. */
	protected double getArrowLength() {
		return 10;
	}

	/** {@inheritDoc} */
	@Override
	public EDragGesture getDragGesture(Node node, DiagramCoordinate diagramLocation) {
		if(node == startFeedbackHandle || node == endFeedbackHandle) {
			return EDragGesture.RECONNECT_LINK;
		}
		for(CurveSegment cs : segments) {
			if(node == cs.getBendPointHandle() || node == cs.getFirstControlPointHandle() ||
					node == cs.getSecondControlPointHandle()) {
				return EDragGesture.MOVE_BENDPOINT;
			}
			if(node == cs.getClickableCurve()) {
				return EDragGesture.NEW_BENDPOINT;
			}
		}
		return EDragGesture.NONE;
	}

	/** {@inheritDoc} */
	@Override
	public void enableFeedback() {
		super.enableFeedback();
		for(CurveSegment cs : segments) {
			cs.setMouseTransparent(false);
		}
	}

	/** {@inheritDoc} */
	@Override
	public void disableFeedback() {
		super.disableFeedback();
		for(CurveSegment cs : segments) {
			cs.setMouseTransparent(true);
		}
	}
}
