/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.rectangular;

import static javafx.scene.paint.Color.TRANSPARENT;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramLayers;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramLayers.ILayer;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.ILinkMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.widgets.LinkArrowWidget;

import javafx.geometry.Point2D;
import javafx.geometry.VPos;
import javafx.scene.paint.Color;
import javafx.scene.paint.Paint;
import javafx.scene.shape.Line;
import javafx.scene.shape.Rectangle;
import javafx.scene.text.Text;
import javafx.scene.text.TextAlignment;

/**
 * This class encapsulates {@link Line}s usually needed for every segment: a visible one, a
 * bigger and invisible one for mouse pointer hit tests. It also provides access to the
 * selection feedback handle of the segment, which is {@code null} for the last segment, since
 * this one ends in an interface node.
 */
final class LineSegment {
	/**
	 * The bend-point model this segments originates from or {@code null} if it is the first
	 * segment.
	 */
	private final Object bendPointModel;
	/** The visible line. */
	private final Line visibleLine;
	/** The invisible line. */
	private final Line clickableLine;
	/** The selection feedback handle rectangle. */
	private final Rectangle feedbackHandle;
	/** The arrow head widget. */
	private final LinkArrowWidget arrowWidget;
	/** Stores the label. */
	private Text label;

	/** Constructor. */
	public LineSegment(double sx, double sy, double ex, double ey, double clickableWidth,
			double markerSize, boolean showArrow, boolean useLineArrow, double arrowLength,
			String labelText) {
		this(null, sx, sy, ex, ey, clickableWidth, markerSize, showArrow, useLineArrow, arrowLength,
				labelText);
	}

	/** Constructor. */
	public LineSegment(Object bendPointModel, double sx, double sy, double ex, double ey,
			double clickableWidth, double feedbackSize, boolean showArrow, boolean useLineArrow,
			double arrowLength, String labelText) {
		this.bendPointModel = bendPointModel;

		this.visibleLine = new Line(sx, sy, ex, ey);

		this.clickableLine = new Line(sx, sy, ex, ey);
		clickableLine.setStroke(TRANSPARENT);
		clickableLine.setStrokeWidth(clickableWidth);

		double fs2 = feedbackSize / 2;
		this.feedbackHandle = new Rectangle(ex - fs2, ey - fs2, feedbackSize, feedbackSize);
		feedbackHandle.setFill(TRANSPARENT);
		feedbackHandle.setStroke(Color.RED);

		if(showArrow) {
			arrowWidget = new LinkArrowWidget(useLineArrow, ex, ey, sx, sy, arrowLength);
		} else {
			arrowWidget = null;
		}
		if(labelText != null) {
			label = new Text(labelText);
			label.setTextAlignment(TextAlignment.CENTER);
			label.setTextOrigin(VPos.TOP);
		} else {
			label = null;
		}
	}

	/** Returns the bend point model. */
	public Object getBendPointModel() {
		return bendPointModel;
	}

	/** Returns the visible line. */
	public Line getVisibleLine() {
		return visibleLine;
	}

	/** Returns the invisible, clickable line. */
	public Line getClickableLine() {
		return clickableLine;
	}

	/** Returns the feedback handle. */
	public Rectangle getFeedbackHandle() {
		return feedbackHandle;
	}

	/** Adds the link nodes of this segment to the link layer node. */
	public void addLinkNodes(DiagramLayers layers, ILinkMVCBundle bundle, String labelText) {
		ILayer visualLayer = layers.getLinkLayer();
		visualLayer.add(visibleLine, bundle);
		if(arrowWidget != null) {
			visualLayer.add(arrowWidget, bundle);
		}
		if(label != null) {
			label.setText(labelText);
			double lx = (visibleLine.getStartX() + visibleLine.getEndX() -
					label.getBoundsInLocal().getWidth()) / 2;
			double ly = (visibleLine.getStartY() + visibleLine.getEndY()) / 2;
			label.setX(lx);
			label.setY(ly);
			layers.getTextLayer().add(label, bundle);
		}
		ILayer interactionLayer = layers.getLinkInteractionLayer();
		interactionLayer.add(clickableLine, bundle);
	}

	/** Removes the link nodes of this segment from the link layer node. */
	public void removeLinkNodes(DiagramLayers layers) {
		ILayer visualLayer = layers.getLinkLayer();
		visualLayer.remove(visibleLine);
		if(arrowWidget != null) {
			visualLayer.remove(arrowWidget);
		}
		if(label != null) {
			layers.getTextLayer().remove(label);
		}
		ILayer interactionLayer = layers.getLinkInteractionLayer();
		interactionLayer.remove(clickableLine);
	}

	/** Adds the feedback nodes of this segment to the feedback layer node. */
	public void addFeedbackNodes(DiagramLayers layers, ILinkMVCBundle bundle) {
		layers.getLinkInteractionLayer().add(feedbackHandle, bundle);
	}

	/** Removes the feedback nodes of this segment from the feedback layer node. */
	public void removeFeedbackNodes(DiagramLayers layers) {
		layers.getLinkInteractionLayer().remove(feedbackHandle);
	}

	/** Sets the color of the visible lines. */
	public void setStrokeColor(Paint color) {
		visibleLine.setStroke(color);
		if(arrowWidget != null) {
			arrowWidget.setColor(color);
		}
	}

	/** Sets the color of the invisible, clickable lines. */
	public void setClickableLinesColor(Paint color) {
		clickableLine.setStroke(color);
	}

	/** Updates the line segment. */
	public void update(Point2D sp, Point2D ep, double feedbackSize, double arrowLength,
			String labelText) {
		double sx = sp.getX();
		double sy = sp.getY();
		visibleLine.setStartX(sx);
		visibleLine.setStartY(sy);
		clickableLine.setStartX(sx);
		clickableLine.setStartY(sy);

		double ex = ep.getX();
		double ey = ep.getY();
		visibleLine.setEndX(ex);
		visibleLine.setEndY(ey);
		clickableLine.setEndX(ex);
		clickableLine.setEndY(ey);

		double fs2 = feedbackSize / 2;
		feedbackHandle.setX(ex - fs2);
		feedbackHandle.setY(ey - fs2);
		feedbackHandle.setWidth(feedbackSize);
		feedbackHandle.setHeight(feedbackSize);

		if(arrowWidget != null) {
			arrowWidget.update(ex, ey, sx, sy, arrowLength);
		}
		if(label != null) {
			label.setText(labelText);
			double lx = (sx + ex - label.getBoundsInLocal().getWidth()) / 2;
			double ly = (sy + ey) / 2;
			label.setX(lx);
			label.setY(ly);
		}
	}

	/** Sets all elements to the given mouse transparency. */
	public void setMouseTransparent(boolean transparent) {
		visibleLine.setMouseTransparent(transparent);
		clickableLine.setMouseTransparent(transparent);
		feedbackHandle.setMouseTransparent(transparent);
		if(arrowWidget != null) {
			arrowWidget.setMouseTransparent(transparent);
		}
		if(label != null) {
			label.setMouseTransparent(transparent);
		}
	}
}
