/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.rectangular;

import static java.lang.Math.hypot;
import static java.util.Objects.requireNonNull;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;

import javafx.geometry.Dimension2D;
import javafx.geometry.Point2D;
import javafx.geometry.Rectangle2D;
import javafx.geometry.Side;

/**
 * This class encapsulates locations on the border of a rectangle defined by the side and offset
 * value. It can be used to compute x-y-locations relative to the upper-left corner of the
 * rectangle.
 * <P>
 * The {@link #getClosestLocationOnBounds(Point2D, Rectangle2D, Dimension2D, Dimension2D)} method
 * allows to construct a {@link RectangularBorderLocation} given a rectangle and a location. The
 * computed location will be on the rectangle and it will be the closest of those locations to the
 * given point.
 */
public final class RectangularBorderLocation {
	/** The side value. */
	private final Side side;
	/** The offset value. */
	private final double offset;
	/** The rectangle used for the computation. */
	private final Rectangle2D rectangle;
	/**
	 * The correction values used to compute the final location (usually half the size of the object
	 * located on the border so that it looks centered.
	 */
	private final Dimension2D anchorageSize;

	/** Constructor. */
	public RectangularBorderLocation(Side side, double offset, Rectangle2D rectangle,
			Dimension2D anchorageSize) {
		this.side = requireNonNull(side);
		this.offset = offset;
		this.rectangle = requireNonNull(rectangle);
		this.anchorageSize = requireNonNull(anchorageSize);
	}

	/** Returns the side. */
	public Side getSide() {
		return side;
	}

	/** Returns the offset. */
	public double getOffset() {
		return offset;
	}

	/**
	 * Returns the location computed from this {@link RectangularBorderLocation} relative to the
	 * upper-left corner of the {@link #rectangle}.
	 */
	public DiagramCoordinate getLocation() {
		if(side == Side.TOP) {
			return new DiagramCoordinate(offset, -anchorageSize.getHeight() / 2);
		}
		if(side == Side.LEFT) {
			return new DiagramCoordinate(-anchorageSize.getWidth() / 2, offset);
		}
		if(side == Side.BOTTOM) {
			return new DiagramCoordinate(offset,
					rectangle.getHeight() - anchorageSize.getHeight() / 2);
		}
		// Side.RIGHT
		return new DiagramCoordinate(rectangle.getWidth() - anchorageSize.getWidth() / 2, offset);
	}

	/**
	 * Returns the closest point on the rectangle bounds when pointing at the given location. Only
	 * locations with distances {@code horizontalSpan} and {@code verticalSpan} are considered.
	 */
	@Deprecated
	public static final RectangularBorderLocation getClosestLocationOnBounds(Point2D p,
			Rectangle2D rectangle, Dimension2D span) {
		LocationOnBorderComputation comp =
				new LocationOnBorderComputation(p, rectangle, span, span);
		RectangularBorderLocation result = new RectangularBorderLocation(comp.getResultSide(),
				comp.getResultOffset(), rectangle, span);
		return result;
	}

	/**
	 * Returns the closest point on the rectangle bounds when pointing at the given location.
	 */
	public static final RectangularBorderLocation getClosestLocationOnBounds(Point2D p,
			Rectangle2D rectangle, Dimension2D anchorageSize, Dimension2D span) {
		LocationOnBorderComputation comp =
				new LocationOnBorderComputation(p, rectangle, anchorageSize, span);
		RectangularBorderLocation result = new RectangularBorderLocation(comp.getResultSide(),
				comp.getResultOffset(), rectangle, anchorageSize);
		return result;
	}

	/** Class for efficiently computing the closest viable position on the parent bounds. */
	private static class LocationOnBorderComputation {
		/** The left-right result offset. */
		private double leftRightOffset;
		/** The left-right result side. */
		private Side leftRightSide;
		/** The left-right distance. */
		private double leftRightDistance;

		/** The top-bottom result offset. */
		private double topBottomOffset;
		/** The top-bottom result side. */
		private Side topBottomSide;
		/** The top-bottom distance. */
		private double topBottomDistance;

		/**
		 * Constructor, which computes the closest location on the given border to the given point.
		 */
		public LocationOnBorderComputation(Point2D pt, Rectangle2D rectangle,
				Dimension2D anchorageSize, Dimension2D snap) {
			// divide the problem by reducing to horizontal and vertical sub-problems
			computeLeftRight(pt, rectangle, anchorageSize.getHeight(), snap.getHeight());
			computeTopBottom(pt, rectangle, anchorageSize.getWidth(), snap.getWidth());
		}

		/** Computes the minimal location on the top or bottom border. */
		private void computeTopBottom(Point2D pt, Rectangle2D rect, double anchorageWidth,
				double snapWidth) {
			// determine top or bottom
			topBottomSide = Side.TOP;
			double borderY = 0;
			if(pt.getY() > rect.getHeight() / 2) {
				topBottomSide = Side.BOTTOM;
				borderY = rect.getHeight();
			}
			// setup helper variables
			double borderWidth = rect.getWidth();
			double correctedX = pt.getX() - anchorageWidth / 2;
			double distanceY = pt.getY() - borderY;

			// scan for closest border location
			double xOffset = 0;
			double currentDistance = hypot(correctedX, distanceY);
			for(double nextOffset = snapWidth; nextOffset < borderWidth; nextOffset += snapWidth) {
				double nextDistance = hypot(correctedX - nextOffset, distanceY);
				if(nextDistance > currentDistance) {
					// distance increased => solution found
					break;
				}
				xOffset = nextOffset;
				currentDistance = nextDistance;
			}
			// record result
			topBottomOffset = xOffset;
			topBottomDistance = currentDistance;
		}

		/** Computes the minimal location on the left or right border. */
		private void computeLeftRight(Point2D pt, Rectangle2D rect, double anchorageHeight,
				double snapHeight) {
			// determine left or right
			leftRightSide = Side.LEFT;
			double borderX = 0;
			if(pt.getX() > rect.getWidth() / 2) {
				leftRightSide = Side.RIGHT;
				borderX = rect.getWidth();
			}
			// setup helper variables
			double borderHeight = rect.getHeight();
			double correctedY = pt.getY() - anchorageHeight / 2;
			double distanceX = pt.getX() - borderX;
			// scan for closest location
			double yOffset = 0;
			double currentDistance = hypot(distanceX, correctedY);
			for(double nextOffset = snapHeight; nextOffset < borderHeight; nextOffset +=
					snapHeight) {
				double nextDistance = hypot(distanceX, correctedY - nextOffset);
				if(nextDistance > currentDistance) {
					// distance increased => solution found
					break;
				}
				yOffset = nextOffset;
				currentDistance = nextDistance;
			}
			// record result
			leftRightOffset = yOffset;
			leftRightDistance = currentDistance;
		}

		/** Returns the side of the result. */
		public Side getResultSide() {
			if(leftRightDistance <= topBottomDistance) {
				return leftRightSide;
			}
			return topBottomSide;
		}

		/** Returns the offset of the result. */
		public double getResultOffset() {
			if(leftRightDistance <= topBottomDistance) {
				return leftRightOffset;
			}
			return topBottomOffset;
		}
	}

	/** {@inheritDoc} */
	@Override
	public String toString() {
		return side.toString() + ":" + offset;
	}
}
