/*******************************************************************************
 * Copyright (c) 2017, 2018 fortiss GmbH. 
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.rectangular;

import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.HIGHLIGHT_INCOMING_LINK_TAG;
import static org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramViewerDefaultTags.HIGHLIGHT_OUTGOING_LINK_TAG;

import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramCoordinate;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.DiagramLayers;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.EDragGesture;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.mvc.IDiagramAnchorageMVCBundle;
import org.eclipse.systemfocus.kernel.common.ui.javafx.lwfxef.visual.base.DiagramAnchorageVisualBase;

import javafx.geometry.Insets;
import javafx.geometry.Rectangle2D;
import javafx.scene.Node;
import javafx.scene.paint.Color;
import javafx.scene.shape.Rectangle;

/**
 * Base class for {@link DiagramAnchorageVisualBase free interface visuals} depicted as rectangles.
 */
public abstract class RectangularDiagramAnchorageVisualBase extends DiagramAnchorageVisualBase {
	/** Constructor. */
	public RectangularDiagramAnchorageVisualBase(IDiagramAnchorageMVCBundle mvcb) {
		super(mvcb, new Rectangle(), new Rectangle());
	}

	/** Returns the visual rectangle. */
	protected final Rectangle getVisualRectangle() {
		// wild cast works: see constructor
		return (Rectangle)getVisualShape();
	}

	/** Returns the hit area rectangle. */
	protected final Rectangle getHitAreaRectangle() {
		// wild cast works: see constructor
		return (Rectangle)getHitAreaShape();
	}

	/** {@inheritDoc} */
	@Override
	public void updateNodes(DiagramLayers layers) {
		super.updateNodes(layers);

		Rectangle2D bounds = getCurrentBounds();
		if(enableVisual()) {
			Rectangle visual = getVisualRectangle();
			Insets i = getInsets();
			visual.setX(bounds.getMinX() + i.getLeft());
			visual.setY(bounds.getMinY() + i.getTop());
			visual.setWidth(bounds.getWidth() - i.getLeft() - i.getRight());
			visual.setHeight(bounds.getHeight() - i.getTop() - i.getBottom());
			visual.setOpacity(getOpacity());
			visual.setStrokeWidth(getBorderWidth());
			visual.setStrokeType(getBorderType());

			if(getMVCBundle().hasTag(HIGHLIGHT_INCOMING_LINK_TAG)) {
				visual.setFill(getHighlightIncomingLinkColor());
				visual.setStroke(getHighlightIncomingLinkBorderColor());
			} else if(getMVCBundle().hasTag(HIGHLIGHT_OUTGOING_LINK_TAG)) {
				visual.setFill(getHighlightOutgoingLinkColor());
				visual.setStroke(getHighlightOutgoingLinkBorderColor());
			} else {
				visual.setFill(getFillColor());
				visual.setStroke(getBorderColor());
			}
		}

		if(enableInteraction()) {
			Rectangle ha = getHitAreaRectangle();
			double hitAreaOutSet = getHitAreaOutset();
			ha.setX(bounds.getMinX() - hitAreaOutSet);
			ha.setY(bounds.getMinY() - hitAreaOutSet);
			ha.setWidth(bounds.getWidth() + 2 * hitAreaOutSet);
			ha.setHeight(bounds.getHeight() + 2 * hitAreaOutSet);
			Color interactionColor = getInteractionColor();
			ha.setStroke(interactionColor);
			ha.setFill(interactionColor);
		}
	}

	/** Returns the insets for the filled rectangle subtracted from {@link #getDimensions()}. */
	protected Insets getInsets() {
		return new Insets(2);
	}

	/** {@inheritDoc} */
	@Override
	protected EDragGesture dragGestureHitTest(Node n, DiagramCoordinate diagramLocation) {
		if(n == getHitAreaRectangle()) {
			Rectangle ha = getHitAreaRectangle();
			double x = diagramLocation.getLocalX(n);
			double y = diagramLocation.getLocalY(n);
			double outset = getHitAreaOutset();
			if(x < outset || x > ha.getWidth() - outset || y < outset ||
					y > ha.getHeight() - outset) {
				return EDragGesture.NEW_LINK;
			}
			return EDragGesture.MOVE;
		}
		if(n == getVisualRectangle()) {
			return EDragGesture.MOVE;
		}
		return EDragGesture.NONE;
	}
}
