/*******************************************************************************
 * Copyright (c) 2016, 2018 fortiss GmbH.
 * 
 * This program and the accompanying materials are made available under the
 * terms of the Eclipse Public License v2.0 which is available at
 * http://www.eclipse.org/legal/epl-v20.html
 * 
 * SPDX-License-Identifier: EPL-2.0
 * 
 * Contributors:
 *     Florian Hoelzl (fortiss GmbH) - initial implementation
 *
 *******************************************************************************/
package org.eclipse.systemfocus.kernel.common.ui.javafx.mesh;

import static javafx.scene.paint.Color.color;
import static org.apache.commons.io.FileUtils.readLines;

import java.io.File;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.systemfocus.kernel.common.ui.javafx.mesh.Mesh3D.IndexPointer;
import org.eclipse.systemfocus.kernel.common.ui.javafx.mesh.Mesh3D.MeshComponent;

import javafx.scene.paint.Color;

/**
 * Utility class for loading OBJ files (i.e. Wavefront format: .obj and .mtl files).
 * These files can be produced by exporting 3D objects from Blender (http://www.blender.org).
 * Note that you must enable the export normals option in order to get the normal vectors, too.
 * Using textures and texture coordinates is optional.
 */
public final class OBJLoader {
	/** Material map. */
	private Map<String, Material3D> materials = new HashMap<String, Material3D>();
	/** Vertex list. */
	private List<Vertex3D> vertices = new ArrayList<Vertex3D>();
	/** Normals list. */
	private List<Normal3D> normals = new ArrayList<Normal3D>();
	/** Texture list. */
	private List<Texture2D> textures = new ArrayList<Texture2D>();
	/** Face list of current object and current material. */
	private List<Face3D> faces = new ArrayList<Face3D>();
	/** Material group list of the current object. */
	private List<MeshComponent> meshComponents = new ArrayList<MeshComponent>();

	/** Material of current object. */
	private Material3D material = null;
	/** Name of the current object. */
	private String objectName = null;
	/** Boolean flag for smooth shading. */
	private boolean smoothShading = false;

	/** The data for the loaded meshes. */
	private List<Mesh3D> meshData;

	/** Constructor for file input with material file derived from this data. */
	public OBJLoader(File objFile) throws IOException {
		List<String> objData = readLines(objFile, "UTF-8");
		for(String line : objData) {
			if(line.startsWith("mtllib ")) { // material library reference found
				File matFile = new File(objFile.getParentFile(), line.split(" ")[1]);
				loadMaterial(readLines(matFile, "UTF-8"));
			}
		}
		loadVerticesTexturesAndNormals(objData);
		loadMeshes(objData);
	}

	/** Constructor for direct object and material content. */
	public OBJLoader(List<String> objectData, List<String> materialData) {
		loadMaterial(materialData);
		loadVerticesTexturesAndNormals(objectData);
		loadMeshes(objectData);
	}

	/** Loads all vertex and normals. */
	private void loadVerticesTexturesAndNormals(List<String> objData) {
		for(String line : objData) {
			if(line.startsWith("#")) { // comment
				continue;
			} else if(line.startsWith("vn ")) {
				addNormal(line.split(" "));
			} else if(line.startsWith("v ")) {
				addVertex(line.split(" "));
			} else if(line.startsWith("vt ")) {
				addTexture(line.split(" "));
			}
		}
	}

	/** Returns the list of loaded meshes. */
	public List<Mesh3D> getLoadedMeshes() {
		return meshData;
	}

	/** Loads the model meshes. */
	private void loadMeshes(List<String> objData) {
		meshData = new LinkedList<Mesh3D>();
		for(String line : objData) {
			if(line.startsWith("#")) { // comment
				continue;
			} else if(line.startsWith("f ")) {
				addFace(line.split(" "));
			} else if(line.startsWith("s ")) {
				if(line.split(" ")[1].equals("off")) {
					smoothShading = false;
				} else {
					smoothShading = true;
				}
			} else if(line.startsWith("o ")) { // next object => new mesh component and mesh
				addMeshComponent();
				addMesh(meshData);
				objectName = line.split(" ")[1];
			} else if(line.startsWith("usemtl ")) { // use material => new mesh component
				addMeshComponent();
				String mat = line.split(" ")[1];
				material = materials.get(mat);
				if(material == null) {
					throw new IllegalArgumentException("Unknown material used in OBJ file: " + mat);
				}
			}
		}
		addMeshComponent();
		addMesh(meshData); // last object;
	}

	/** Adds a mesh component and clears temporary data. */
	private void addMeshComponent() {
		if(material != null) {
			meshComponents.add(new MeshComponent(material, faces, smoothShading));
			material = null;
			smoothShading = false;
			faces.clear();
		}
	}

	/** Adds an object if some faces exists and clear temporary data. */
	private void addMesh(List<Mesh3D> result) {
		if(!meshComponents.isEmpty()) {
			result.add(new Mesh3D(objectName, vertices, textures, normals, meshComponents));
			objectName = null;
			material = null;
			faces.clear();
			meshComponents.clear();
		}
	}

	/** Adds a face. */
	private void addFace(String[] tokens) {
		if(tokens.length != 4) {
			throw new IllegalArgumentException("OBJ file contains non-triangles.");
		}
		IndexPointer v0 = getVertexData(tokens[1]);
		IndexPointer v1 = getVertexData(tokens[2]);
		IndexPointer v2 = getVertexData(tokens[3]);
		faces.add(new Face3D(v0, v1, v2));
	}

	/** Returns the indices in the vertex and normals list. Texture is currently ignored. */
	private IndexPointer getVertexData(String s) {
		String[] ids = s.split("/");
		if(ids.length != 3) {
			throw new IllegalArgumentException("OBJ file does not contain normal vectors.");
		}
		// Indices in OBJ files have base 1, not 0!
		int vertexIdx = "".equals(ids[0]) ? 0 : Integer.valueOf(ids[0]) - 1;
		int textureIdx = "".equals(ids[1]) ? 0 : Integer.valueOf(ids[1]) - 1;
		int normalIdx = "".equals(ids[2]) ? 0 : Integer.valueOf(ids[2]) - 1;
		return new IndexPointer(vertexIdx, textureIdx, normalIdx);
	}

	/** Adds a vertex. */
	private void addVertex(String[] tokens) {
		vertices.add(new Vertex3D(Float.valueOf(tokens[1]), Float.valueOf(tokens[2]),
				Float.valueOf(tokens[3])));
	}

	/** Adds a normal. */
	private void addNormal(String[] tokens) {
		normals.add(new Normal3D(Float.valueOf(tokens[1]), Float.valueOf(tokens[2]),
				Float.valueOf(tokens[3])));
	}

	/** Adds a texture coordinate. */
	private void addTexture(String[] tokens) {
		textures.add(new Texture2D(Float.valueOf(tokens[1]), Float.valueOf(tokens[2])));
	}

	/** Loads the materials from the given URL and stores them in the map. */
	public void loadMaterial(List<String> lines) {
		String matName = null;
		Color ambient = null;
		Color diffuse = null;
		Color specular = null;
		float specularCoeff = 0;

		for(String line : lines) {
			if(line.startsWith("#")) { // comment
				continue;
			}
			String[] tokens = line.split(" ");
			if(tokens.length > 1) {
				if("newmtl".equals(tokens[0])) {
					if(matName != null) {
						Material3D newMat =
								new Material3D(ambient, diffuse, specular, specularCoeff);
						materials.put(matName, newMat);
						ambient = color(0.0f, 0.0f, 0.0f);
						diffuse = color(0.0f, 0.0f, 0.0f);
						specular = color(0.0f, 0.0f, 0.0f);
						specularCoeff = 0;
					}
					matName = tokens[1];
				} else if("Ka".equals(tokens[0])) {
					ambient = color(Float.valueOf(tokens[1]), Float.valueOf(tokens[2]),
							Float.valueOf(tokens[3]));
				} else if("Kd".equals(tokens[0])) {
					diffuse = color(Float.valueOf(tokens[1]), Float.valueOf(tokens[2]),
							Float.valueOf(tokens[3]));
				} else if("Ks".equals(tokens[0])) {
					specular = color(Float.valueOf(tokens[1]), Float.valueOf(tokens[2]),
							Float.valueOf(tokens[3]));
				} else if("Ns".equals(tokens[0])) {
					specularCoeff = Float.valueOf(tokens[1]);
				}
			}
		}

		if(matName != null) {
			Material3D newMat = new Material3D(ambient, diffuse, specular, specularCoeff);
			materials.put(matName, newMat);
		}
	}
}
